/* $Header: /home/yav/catty/lupe/RCS/lupe.c,v 1.9 1995/06/02 01:32:41 yav Exp $
 * Lupe - real-time magnifying glass for X11
 * written by yav (UHD98984@pcvan.or.jp)
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <stdio.h>
#include "config.h"

#ifdef X11_EXTENSIONS_SHAPE_H
#include <X11/extensions/shape.h>
#endif

#include "version.h"

char *str_display = "";
char *str_geometry = "";
char *str_border = NULL;	/* border color */
int magstep = 6;		/* magnify factor */
int bw = 2;			/* border width */
int update_time = 333;		/* 1/3 sec. */
int delay_time = 5000;		/* 5 sec. */
int shape_mode = 1;		/* != 0  use shape extension */
int reticle_mode = 1;		/* != 0  display reticle */
static char *str_lupe = "lupe";
static Display *dsp;
static int scr;
static Window rootwin;
static Window topwin;
static int srcx, srcy;
static int srcw = 25;		/* source width */
static int srch = 25;		/* source height */
static int dstw;		/* destination width = srcw * magstep */
static int dsth;		/* destination height = srch * magstep */
static Pixmap srcpixmap = None;
static Pixmap dstpixmap = None;
static GC gc;
static GC linegc;
static int oargc;
static char **oargv;
static Bool mapped = False;

#include "icon.xbm"

/* option.c */
extern int parse_option();
/* timer.c */
extern void reset_time();
extern long elapsed_time();
extern void sleep_time();

void usage()
{
  fprintf(stderr, "* Lupe version %s *\n", LUPE_VERSION);
  fprintf(stderr, "written by yav (UHD98984@pcvan.or.jp)\n");
  fprintf(stderr, "usage: %s [-display DISP] [-geometry WxH+X+Y] [-bw N]\n",
	  *oargv);
  fprintf(stderr, "  [-mag N] [-shape|-noshape] [-reticle|-noreticle]\n");
  fprintf(stderr, "  [-update milliseconds] [-delay milliseconds]\n");
  exit(1);
}

void expose(x, y, w, h)
     int x;
     int y;
     int w;
     int h;
{
  XCopyArea(dsp, dstpixmap, topwin, gc, x, y, w, h, x, y);
}

void magnify()
{
  int i, x, y, x2, y2;
  
  for (x = 0; x < srcw; x++) {
    x2 = x*magstep;
    for (i = 0; i < magstep; i++) {
      XCopyArea(dsp, srcpixmap, dstpixmap, gc, x, 0, 1, srch, x2+i, 0);
    }
  }
  x2 = dstw;
  for (y = srch-1; y >= 0; --y) {
    y2 = y*magstep;
    for (i = magstep-1; i >= 0; --i) {
      XCopyArea(dsp, dstpixmap, dstpixmap, gc, 0, y, x2, 1, 0, y2+i);
    }
  }
  if (reticle_mode) {
    /* vertical line */
    x = (srcw/2)*magstep + magstep/2;
    XDrawLine(dsp, dstpixmap, linegc, x, (srch/2)*magstep-1, x, dsth/4);
    XDrawLine(dsp, dstpixmap, linegc, x, (srch/2+1)*magstep, x, dsth*3/4);
    /* horizontal line */
    y = (srch/2)*magstep + magstep/2;
    XDrawLine(dsp, dstpixmap, linegc, (srcw/2)*magstep-1, y, dstw/4, y);
    XDrawLine(dsp, dstpixmap, linegc, (srcw/2+1)*magstep, y, dstw*3/4, y);
  }
}

int get_image()
{
  int x, y;
  int rootx, rooty;
  unsigned int kb;
  Window root, child;
  int pointer_move;
  
  XQueryPointer(dsp, rootwin, &root, &child, &rootx, &rooty, &x, &y, &kb);
  pointer_move = x != srcx || y != srcy;
  srcx = x;
  srcy = y;
  XFillRectangle(dsp, srcpixmap, gc, 0, 0, srcw, srch);
  XCopyArea(dsp, rootwin, srcpixmap, gc, x-srcw/2, y-srch/2, srcw, srch, 0, 0);
  magnify();
  expose(0, 0, dstw, dsth);
  return pointer_move;
}

void set_shape()
{
#ifdef X11_EXTENSIONS_SHAPE_H
  Pixmap bitmap;
  GC gc;
  
  bitmap = XCreatePixmap(dsp, rootwin, dstw, dsth, 1);
  gc = XCreateGC(dsp, bitmap, 0, NULL);
  XSetForeground(dsp, gc, 0);
  XFillRectangle(dsp, bitmap, gc, 0, 0, dstw, dsth);
  XSetForeground(dsp, gc, 1);
  XFillArc(dsp, bitmap, gc, bw, bw, dstw-bw*2, dsth-bw*2, 0, 360*64);
  XShapeCombineMask(dsp, topwin, ShapeClip, 0, 0, bitmap, ShapeSet);
  XFillArc(dsp, bitmap, gc, 0, 0, dstw, dsth, 0, 360*64);
  XShapeCombineMask(dsp, topwin, ShapeBounding, 0, 0, bitmap, ShapeSet);
  XFreeGC(dsp, gc);
  XFreePixmap(dsp, bitmap);
#endif
}

void resize()
{
  srcw = dstw / magstep;
  srch = dsth / magstep;
  if (srcpixmap != None)
    XFreePixmap(dsp, srcpixmap);
  srcpixmap = XCreatePixmap(dsp, rootwin, srcw, srch, DefaultDepth(dsp, scr));
  if (dstpixmap != None)
    XFreePixmap(dsp, dstpixmap);
  dstpixmap = XCreatePixmap(dsp, rootwin, dstw, dsth, DefaultDepth(dsp, scr));
  XFillRectangle(dsp, dstpixmap, gc, 0, 0, dstw, dsth);
  if (shape_mode)
    set_shape();
}

void init_screen()
{
  int f, x, y, w, h;
  XSizeHints hint;
  unsigned long border_pixel;
  XColor col;
  Pixmap icon_pixmap;
  static char dash_list[2] = {1,1};
  
  dsp = XOpenDisplay(str_display);
  if (dsp == NULL) {
    fprintf(stderr, "Cannot open display connection!\n");
    exit(1);
  }
  scr = DefaultScreen(dsp);
  rootwin = DefaultRootWindow(dsp);
  border_pixel = BlackPixel(dsp, scr);
  if (str_border != NULL &&
      XParseColor(dsp, DefaultColormap(dsp, scr), str_border, &col) &&
      XAllocColor(dsp, DefaultColormap(dsp, scr), &col))
    border_pixel = col.pixel;
  icon_pixmap = 
    XCreateBitmapFromData(dsp, DefaultRootWindow(dsp), 
			  icon_bits, icon_width, icon_height);
  hint.min_width = hint.min_height =
    hint.width_inc = hint.height_inc = magstep;
  hint.flags = PMinSize|PResizeInc;
  f = XGeometry(dsp, scr, str_geometry, "",
		bw, magstep, magstep, 0, 0, &x, &y, &w, &h);
  if (f & WidthValue) {
    hint.flags |= USSize;
    srcw = w;
  }
  if (f & HeightValue) {
    hint.flags |= USSize;
    srch = h;
  }
  if (f & XValue)
    hint.flags |= USPosition;
  else
    x = 0;
  if (f & YValue)
    hint.flags |= USPosition;
  else
    y = 0;
  dstw = srcw * magstep;
  dsth = srch * magstep;
  hint.x = x;
  hint.y = y;
  hint.width = dstw;
  hint.height = dsth;
  topwin = XCreateSimpleWindow(dsp, rootwin,
			       hint.x, hint.y, hint.width, hint.height,
			       bw, border_pixel,
			       WhitePixel(dsp, DefaultScreen(dsp)));
  XSetWindowBorderPixmap(dsp, topwin, CopyFromParent);
  XSetWindowBackgroundPixmap(dsp, topwin, None);
  XSelectInput(dsp, topwin, ExposureMask|ButtonPressMask|StructureNotifyMask);
  XSetStandardProperties(dsp, topwin, str_lupe, str_lupe, icon_pixmap,
			 oargv, oargc, &hint);
  /* setup GC */
  gc = XCreateGC(dsp, rootwin, 0, NULL);
  XSetSubwindowMode(dsp, gc, IncludeInferiors);
  linegc = XCreateGC(dsp, rootwin, 0, NULL);
  XSetLineAttributes(dsp, linegc, 0, LineDoubleDash, CapButt, JoinMiter);
  XSetDashes(dsp, linegc, 0, dash_list, sizeof(dash_list));
  resize();
  XMapRaised(dsp, topwin);
}

void adjust_option()
{
  if (magstep < 1)
    magstep = 1;
  if (srcw < 1)
    srcw = 1;
  if (srch < 1)
    srch = 1;
}

void main(argc, argv)
     int argc;
     char **argv;
{
  Bool fin;
  XEvent ev;

  oargc = argc--;
  oargv = argv++;
  while (argc > 0) {
    if (parse_option(&argc, &argv)) {
      fprintf(stderr, "%s: unknown option ``%s''\n", *oargv, *argv);
      usage();
    }
  }
  adjust_option();
  init_screen();
  fin = False;
  reset_time();
  while (!fin) {
    while (XEventsQueued(dsp, QueuedAfterReading) == 0) {
      XFlush(dsp);
      if (elapsed_time() >= delay_time)
	sleep_time(update_time);
      if (mapped)
	get_image();
    }
    XNextEvent(dsp, &ev);
    switch (ev.type) {
    case MotionNotify:
      while (XCheckTypedEvent(dsp, MotionNotify, &ev))
	;
      get_image();
      reset_time();
      break;
    case Expose:
      if (ev.xexpose.count == 0)
	expose(ev.xexpose.x, ev.xexpose.y, ev.xexpose.width, ev.xexpose.height);
      break;
    case ButtonPress:
      switch(ev.xbutton.button) {
      case Button1:
	XIconifyWindow(dsp, topwin, DefaultScreen(dsp));
	break;
      case Button2:
      case Button3:
	fin = True;
	break;
      }
      break;
    case MapNotify:
      XSelectInput(dsp, rootwin, PointerMotionHintMask|PointerMotionMask);
      mapped = True;
      break;
    case UnmapNotify:
      XSelectInput(dsp, rootwin, 0);
      mapped = False;
      break;
    case ConfigureNotify:
      dstw = ev.xconfigure.width;
      dsth = ev.xconfigure.height;
      resize();
      break;
    }
  }
  XCloseDisplay(dsp);
  exit(0);
}

/* End of file */
